from AccessControl import getSecurityManager
from AccessControl import Unauthorized
from Products.PageTemplates.PageTemplateFile import PageTemplateFile as Z2PTF
from zope.pagetemplate.pagetemplatefile import PageTemplateFile as Z3PTF
from Products.Five.browser.pagetemplatefile import BoundPageTemplate as Z2BPT


def isTemplate(obj):
    """ check if the given object is a or is derived from a template class """
    # TODO: we should really check via interfaces, i.e. `providedBy` here,
    #       but the only class using interfaces atm is Z3PTF :(
    return isinstance(obj, (Z2PTF, Z3PTF, Z2BPT))


def findViewletTemplate(viewlet):
    """ try to find the attribute holding the template within a viewlet """
    for attr in 'index', 'template', '_template', '__call__', 'render':
        item = getattr(viewlet, attr, None)
        if isTemplate(item):
            return attr, item
    attrs = [ attr for attr in dir(viewlet) if isTemplate(getattr(viewlet, attr, None)) ]
    if len(attrs) == 1:
        return attrs[0], getattr(viewlet, attrs[0])
    else:
        # TODO: we should pass on the message if we find multiple templates
        pass
    return None, None


def checkPermission(permission, context):
    sm = getSecurityManager()
    if permission is not None:
        if not sm.checkPermission(permission, context):
            raise Unauthorized('The current user does not have the '
                               'required "%s" permission' % permission)




# -*- coding: utf-8 -*-
from Acquisition import aq_inner
from Products.Five.browser import BrowserView
from Products.Five.component import findSite
from Products.Five.component.interfaces import IObjectManagerSite
from five.customerize.interfaces import IViewTemplateContainer
from five.customerize.zpt import TTWViewTemplate
from os.path import sep, isabs, split, basename
from zope.component import getGlobalSiteManager
from zope.component import getMultiAdapter, getSiteManager
from zope.component import getUtility, queryUtility
from zope.dottedname.resolve import resolve
from zope.interface import providedBy, Interface
from zope.interface.interfaces import IInterface
from zope.publisher.interfaces import IRequest
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.schema.interfaces import IVocabularyFactory
from zope.traversing.browser import absoluteURL

import six


# This method was copied from zope.app.apidoc.presentation
def getViews(iface, type=IRequest):
    """Get all view registrations for a particular interface."""
    gsm = getGlobalSiteManager()
    for reg in gsm.registeredAdapters():
        if (len(reg.required) > 0 and
            reg.required[-1] is not None and
            reg.required[-1].isOrExtends(type)):

            for required_iface in reg.required[:-1]:
                if required_iface is None or iface.isOrExtends(required_iface):
                    yield reg


def mangleAbsoluteFilename(filename):
    """
    Mangle an absolute filename when the file happens to be in a
    package.  The mangled name will then be of the form::

      <dotted package name>/<base filename>.

    For example, let's take Five's configure.zcml as an example.  We
    assemble it in an OS-independent way so this test works on all
    platforms:

      >>> def filesystemPath(*elements):
      ...     return sep.join(elements)

    We see that the filename is now mangled:

      >>> mangleAbsoluteFilename(filesystemPath(
      ...     '', 'path', 'to', 'Products', 'Five', 'configure.zcml'))
      'Products.Five/configure.zcml'

    The name of a file that's not in a package is returned unchanged:

      >>> not_in_a_package = filesystemPath('', 'path', 'to', 'configure.zcml')
      >>> mangleAbsoluteFilename(not_in_a_package) == not_in_a_package
      True
    """
    if not isabs(filename):
        raise ValueError("Can only determine package for absolute filenames")
    dir, base = split(filename)
    pieces = dir.split(sep)
    if pieces[0] == '':
        pieces = pieces[1:]

    # get pieces that comes after the last folder whose name starts by a dot
    for index, piece in enumerate(reversed(pieces)):
        if piece.startswith('.'):
            pieces = pieces[-index:]
            break

    while pieces:
        try:
            resolve('.'.join(pieces))
            break
        except (ImportError, ValueError):
            pieces = pieces[1:]
    if not pieces:
        return filename
    return '.'.join(pieces) + '/' + base

class CustomizationView(BrowserView):

    def templateViewRegistrations(self):
        for reg in getViews(providedBy(self.context), IBrowserRequest):
            factory = reg.factory
            while hasattr(factory, 'factory'):
                factory = factory.factory
            #XXX this should really be dealt with using a marker interface
            # on the view factory
            if hasattr(factory, '__name__') and \
                   factory.__name__.startswith('SimpleViewClass'):
                yield reg

    def templateViewRegInfo(self):
        def regkey(reg):
            return reg.name
        for reg in sorted(self.templateViewRegistrations(), key=regkey):
            yield {
                'viewname': reg.name,
                'for': reg.required[0].__identifier__,
                'type': reg.required[1].__identifier__,
                'zptfile': mangleAbsoluteFilename(reg.factory.index.filename),
                'zcmlfile': mangleAbsoluteFilename(reg.info.file)
                }

    def viewClassFromViewName(self, viewname):
        view = getMultiAdapter((self.context, self.request), name=viewname)
        # The view class is generally auto-generated, we usually want
        # the first base class, though if the view only has one base
        # (generally object or BrowserView) we return the full class
        # and hope that it can be pickled
        klass = view.__class__
        base = klass.__bases__[0]
        if base is BrowserView or base is object:
            return klass
        return base

    def templateFromViewName(self, viewname):
        view = getMultiAdapter((self.context, self.request), name=viewname)
        return view.index

    def templateCodeFromViewName(self, viewname):
        template = self.templateFromViewName(viewname)
        #XXX: we can't do template.read() here because of a bug in
        # Zope 3's ZPT implementation.
        data = open(template.filename, 'rb').read()
        if six.PY2:
            return data
        return data.decode('utf-8')

    def permissionFromViewName(self, viewname):
        view = getMultiAdapter((self.context, self.request), name=viewname)
        permissions = view.__class__.__ac_permissions__
        for permission, methods in permissions:
            if methods[0] in ('', '__call__'):
                return permission

    def doCustomizeTemplate(self, viewname):
        # find the nearest site
        site = findSite(self.context, IObjectManagerSite)
        if site is None:
            raise TypeError("No site found")  # TODO find right exception

        # we're using the original filename of the template, not the
        # view name to avoid potential conflicts and/or confusion in
        # URLs
        template = self.templateFromViewName(viewname)
        zpt_id = basename(template.filename)

        template_file = self.templateCodeFromViewName(viewname)
        viewclass = self.viewClassFromViewName(viewname)
        permission = self.permissionFromViewName(viewname)
        viewzpt = TTWViewTemplate(zpt_id, template_file, view=viewclass,
                                  permission=permission)
        container = queryUtility(IViewTemplateContainer)
        if container is not None:
            viewzpt = container.addTemplate(zpt_id, viewzpt)
        else:
            site._setObject(zpt_id, viewzpt) #XXXthere could be a naming conflict
            viewzpt = getattr(site, zpt_id)

        # find out the view registration object so we can get at the
        # provided and required interfaces
        for reg in getViews(providedBy(self.context), IBrowserRequest):
            if reg.name == viewname:
                break

        components = site.getSiteManager()
        components.registerAdapter(viewzpt, required=reg.required,
                                   provided=reg.provided, name=viewname
                                   ) #XXX info?

        return viewzpt

    def customizeTemplate(self, viewname):
        viewzpt = self.doCustomizeTemplate(viewname)
        # to get a "direct" URL we use aq_inner for a straight
        # acquisition chain
        url = absoluteURL(aq_inner(viewzpt), self.request) + "/manage_workspace"
        self.request.response.redirect(url)

class RegistrationsView(BrowserView):

    def viewRegistrations(self):
        regs = []
        components = getSiteManager(self.context)
        for reg in components.registeredAdapters():
            if (len(reg.required) == 2 and
                reg.required[1].isOrExtends(IBrowserRequest) and
                reg.factory == self.context):
                regs.append(reg)
        def regkey(reg):
            return (reg.name, reg.required)
        return sorted(regs, key=regkey)

    def getAllInterfaceNames(self):
        factory = getUtility(IVocabularyFactory, 'Interfaces')
        vocab = factory(self.context)
        return (term.token for term in vocab)

    def getRequestInterfaceNames(self):
        factory = getUtility(IVocabularyFactory, 'Interfaces')
        vocab = factory(self.context)
        return (term.token for term in vocab
                if term.value.isOrExtends(IBrowserRequest))

    # TODO needs tests
    def unregister(self):
        index = self.request.form.get('index')
        try:
            index = int(index)
        except (TypeError, ValueError):
            index = None
        if index is None:
            #XXX: find right exception type
            raise ValueError("Missing or invalid 'index' parameter.")
        regs = self.viewRegistrations()
        reg = regs[index]
        components = getSiteManager(self.context)
        components.unregisterAdapter(reg.factory, reg.required, reg.provided,
                                     reg.name)
        self.request.response.redirect('registrations.html')

    # TODO needs tests
    def register(self, for_name, type_name, name='', comment=''):
        for_ = getUtility(IInterface, for_name)
        type = getUtility(IInterface, type_name)
        components = getSiteManager(self.context)
        components.registerAdapter(self.context, (for_, type),
                                   Interface, name, comment)
        self.request.response.redirect('registrations.html')



from zope.component import adapter, getSiteManager
from zope.viewlet.viewlet import ViewletBase
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.viewlet.interfaces import IViewlet, IViewletManager
from zope.interface import implementer

from five.customerize.interfaces import ITTWViewTemplate
from five.customerize.utils import checkPermission
from plone.portlets.interfaces import IPortletRenderer
from plone.portlets.interfaces import IPortletManager


@implementer(ITTWViewTemplate)
class TTWViewTemplate(ZopePageTemplate):
    """A template class used to generate Zope 3 views TTW"""

    manage_options = (
        ZopePageTemplate.manage_options[0],
        dict(label='Registrations', action='registrations.html'),
        ) + ZopePageTemplate.manage_options[2:]

    def __init__(self, id, text=None, content_type='text/html', strict=True,
                 encoding='utf-8', view=None, permission=None, name=None):
        self.view = view
        self.permission = permission
        self.name = name
        super(TTWViewTemplate, self).__init__(id, text, content_type, encoding,
                                              strict)

    def __call__(self, context, request, view=None, manager=None, data=None):
        #XXX raise a sensible exception if context and request are
        # omitted, IOW, if someone tries to render the template not as
        # a view.

        # the security check is now deferred until the template/viewlet/portlet
        # is actually called, because it may be looked up during traversal,
        # in which case there's no proper security context yet
        if IPortletManager.providedBy(manager):
            return TTWPortletRenderer(context, request, self, view,
                manager, data, self.permission)
        if IViewletManager.providedBy(manager):
            return TTWViewletRenderer(context, request, self, view,
                manager, self.permission)
        else:
            return TTWViewTemplateRenderer(context, request, self, self.permission)

    # overwrite Shared.DC.Scripts.Binding.Binding's before traversal
    # hook that would prevent to look up views for instances of this
    # class.
    def __before_publishing_traverse__(self, self2, request):
        pass


class TTWViewTemplateRenderer(object):
    """The view object for the TTW View Template.

    When a TTWViewTemplate-based view is looked up, an object of this
    class is instantiated.  It holds a reference to the
    TTWViewTemplate object which it will use in the render process
    (__call__).
    """

    def __init__(self, context, request, template, permission=None):
        self.context = context
        self.request = request
        self.template = template
        self.view = None
        self.permission = permission

    def __call__(self, *args, **kwargs):
        """Render the TTWViewTemplate-based view.
        """
        view = self._getView()
        # we need to override the template's context and request as
        # they generally point to the wrong objects (a template's
        # context usually is what it was acquired from, which isn't
        # what the context is for a view template).
        bound_names = {'context': self.context,
                       'request': self.request,
                       'view': view}
        template = self.template.__of__(self.context)
        return template._exec(bound_names, args, kwargs)

    def _getView(self):
        checkPermission(self.permission, self.context)
        if self.view is not None:
            return self.view
        view_class = self.template.view
        if view_class is not None:
            # Filesystem-based view templates are trusted code and
            # have unrestricted access to the view class.  We simulate
            # that for TTW templates (which are trusted code) by
            # creating a subclass with unrestricted access to all
            # subobjects.
            if hasattr(view_class, '_five_customerize_ttw_class'):
                TTWView = view_class._five_customerize_ttw_class
            else:
                class TTWView(view_class):
                    __allow_access_to_unprotected_subobjects__ = 1
                view_class._five_customerize_ttw_class = TTWView
            self.view = TTWView(self.context, self.request)
            return self.view

    # Zope 2 wants to acquisition-wrap every view object (via __of__).
    # We don't need this as the TTWViewTemplate object is already
    # properly acquisition-wrapped in __call__.  Nevertheless we need
    # to support the __of__ method as a no-op.
    def __of__(self, obj):
        return self

    # Make sure this object passes the ZPublisher's security validation
    @property
    def __parent__(self):
        return self.context


@implementer(IViewlet)
class TTWViewletRenderer(object):
    """ analogon to TTWViewTemplateRenderer for viewlets """

    __allow_access_to_unprotected_subobjects__ = True

    def __init__(self, context, request, template, view, manager=None, permission=None):
        self.context = context
        self.request = request
        self.template = template
        self.view = view
        self.manager = manager
        self.viewlet = None
        self.permission = permission

    def update(self):
        """ update the viewlet before `render` is called """
        self._getViewlet().update()

    def render(self, *args, **kwargs):
        """ render the viewlet using the customized template """
        viewlet = self._getViewlet()
        # we need to override the template's context and request as
        # they generally point to the wrong objects (a template's
        # context usually is what it was acquired from, which isn't
        # what the context is for a view template).
        bound_names = {'context': self.context,
                       'request': self.request,
                       'view': viewlet}
        template = self.template.__of__(self.context)
        return template._exec(bound_names, args, kwargs)

    def _getViewlet(self):
        checkPermission(self.permission, self.context)
        if self.viewlet is not None:
            return self.viewlet
        view_class = self.template.view
        if view_class is not None:
            # Filesystem-based view templates are trusted code and
            # have unrestricted access to the view class.  We simulate
            # that for TTW templates (which are trusted code) by
            # creating a subclass with unrestricted access to all
            # subobjects.
            if hasattr(view_class, '_five_customerize_ttw_class'):
                TTWViewlet = view_class._five_customerize_ttw_class
            else:
                class TTWViewlet(view_class, ViewletBase):
                    __allow_access_to_unprotected_subobjects__ = 1
                view_class._five_customerize_ttw_class = TTWViewlet
            self.viewlet = TTWViewlet(self.context, self.request, self.view, self.manager)
            return self.viewlet

    # Zope 2 wants to acquisition-wrap every view object (via __of__).
    # We don't need this as the TTWViewTemplate object is already
    # properly acquisition-wrapped in __call__.  Nevertheless we need
    # to support the __of__ method as a no-op.
    def __of__(self, obj):
        return self


@implementer(IPortletRenderer)
class TTWPortletRenderer(object):
    """ analogon to TTWViewletRenderer for portlets """

    __allow_access_to_unprotected_subobjects__ = True

    def __init__(self, context, request, template, view, manager=None, data=None, permission=None):
        self.context = context
        self.request = request
        self.template = template
        self.view = view
        self.manager = manager
        self.data = data
        self.renderer = None
        self.permission = permission

    def update(self):
        """ update the portlet before `render` is called """
        self._getRenderer().update()

    def render(self, *args, **kwargs):
        """ render the portlet using the customized template """
        renderer = self._getRenderer()
        # we need to override the template's context and request as
        # they generally point to the wrong objects (a template's
        # context usually is what it was acquired from, which isn't
        # what the context is for a view template).
        bound_names = {'context': self.context,
                       'request': self.request,
                       'view': renderer}
        template = self.template.__of__(self.context)
        return template._exec(bound_names, args, kwargs)

    def _getRenderer(self):
        checkPermission(self.permission, self.context)
        if self.renderer is not None:
            return self.renderer
        view_class = self.template.view
        if view_class is not None:
            # Filesystem-based view templates are trusted code and
            # have unrestricted access to the view class.  We simulate
            # that for TTW templates (which are trusted code) by
            # creating a subclass with unrestricted access to all
            # subobjects.
            if hasattr(view_class, '_five_customerize_ttw_class'):
                TTWPortlet = view_class._five_customerize_ttw_class
            else:
                class TTWPortlet(view_class):
                    __allow_access_to_unprotected_subobjects__ = 1
                view_class._five_customerize_ttw_class = TTWPortlet
            self.renderer = TTWPortlet(self.context, self.request, self.view, self.manager, self.data)
            return self.renderer

    @property
    def available(self):
        return self._getRenderer().available

    # Zope 2 wants to acquisition-wrap every view object (via __of__).
    # We don't need this as the TTWViewTemplate object is already
    # properly acquisition-wrapped in __call__.  Nevertheless we need
    # to support the __of__ method as a no-op.
    def __of__(self, obj):
        return self


@adapter(TTWViewTemplate, IObjectRemovedEvent)
def unregisterViewWhenZPTIsDeleted(zpt, event):
    components = getSiteManager(zpt)
    for reg in components.registeredAdapters():
        if reg.factory == zpt:
            components.unregisterAdapter(reg.factory, reg.required,
                                         reg.provided, reg.name)
            break



from zope.interface import Interface


class ITTWViewTemplate(Interface):
    """ ttw customizable page template view """

    def __call__(context, request):
        """ render the template/view """


class IViewTemplateContainer(Interface):
    """ container for all ttw view template objects """

    def addTemplate(id, template):
        """ add the given ttw view template to the container
            and return it acquisition wrapped in the container """




from plone.testing import Layer
from plone.testing import layered
from plone.testing import zca
from plone.testing import zope
from Products.Five.browser import BrowserView
from zope.configuration import xmlconfig

import doctest
import re
import six
import unittest


class FiveCustomerizeLayer(Layer):
    defaultBases = (zope.STARTUP,)

    def setUp(self):
        # Stack a new configuration context
        self['configurationContext'] = context = zca.stackConfigurationContext(
            self.get('configurationContext'))

        import Products.Five
        import five.customerize
        xmlconfig.file('configure.zcml', Products.Five, context=context)
        xmlconfig.file('configure.zcml', five.customerize, context=context)

    def tearDown(self):
        # Zap the stacked configuration context
        del self['configurationContext']


FIVE_CUSTOMERIZE_FIXTURE = FiveCustomerizeLayer()

FIVE_CUSTOMERIZE_FUNCTIONAL_TESTING = zope.FunctionalTesting(
    bases=(FIVE_CUSTOMERIZE_FIXTURE,), name="five.customerize:FUNCTIONAL")


class Py23DocChecker(doctest.OutputChecker):
    def check_output(self, want, got, optionflags):
        if six.PY2:
            got = re.sub('Unauthorized', 'AccessControl.unauthorized.Unauthorized', got)
            got = re.sub("u'(.*?)'", "'\\1'", got)
        return doctest.OutputChecker.check_output(self, want, got, optionflags)


class TestView(BrowserView):
    """A view class"""
    __name__ = 'mystaticview.html'

    def foo_method(self):
        return 'baz'

    def __call__(self):
        return 'Original View'


def test_suite():
    return unittest.TestSuite([
        layered(
            doctest.DocFileSuite('zpt.txt'),
            layer=FIVE_CUSTOMERIZE_FUNCTIONAL_TESTING,
        ),
        layered(
            doctest.DocFileSuite(
                'customerize.txt',
                checker=Py23DocChecker(),
            ),
            layer=FIVE_CUSTOMERIZE_FUNCTIONAL_TESTING,
        ),
        layered(
            doctest.DocFileSuite('browser.txt', globs={
                'TestView': TestView,
            }),
            layer=FIVE_CUSTOMERIZE_FUNCTIONAL_TESTING,
        ),
    ])



##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from setuptools import setup

__version__ = '2.0.1'

setup(
    name='five.customerize',
    version=__version__,
    description='TTW customization of template-based Zope browser views',
    long_description=(open('README.rst').read() + "\n" +
                      open('CHANGES.rst').read()),
    keywords='zope views templates customization ttw',
    author='Zope Foundation and Contributors',
    author_email='zope-dev@zope.org',
    url='https://github.com/zopefoundation/five.customerize',
    license='ZPL 2.1',
    packages=['five', 'five.customerize'],
    package_dir={'': 'src'},
    namespace_packages=['five'],
    include_package_data=True,
    platforms='Any',
    zip_safe=False,
    install_requires=[
        'setuptools',
        'plone.portlets',
        'zope.component',
        'zope.componentvocabulary',
        'zope.dottedname',
        'zope.interface',
        'zope.lifecycleevent',
        'zope.pagetemplate',
        'zope.schema',
        'zope.traversing',
        'zope.viewlet',
        'Acquisition',
        'Zope',
    ],
    extras_require={
        'test': [
            'plone.testing>=7',
            'zope.publisher',
            'zope.site',
            'zope.testing',
            'transaction',
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Zope :: 4',
        'License :: OSI Approved :: Zope Public License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
    ],
)
